# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import tempfile
from dataclasses import dataclass
from unittest import mock
from unittest.mock import AsyncMock

import pytest

from airflow.providers.smtp.notifications.smtp import SmtpNotifier, send_smtp_notification

from tests_common.test_utils.version_compat import AIRFLOW_V_3_1_PLUS

TRY_NUMBER = 0

SMTP_CONN_ID = "smtp_default"
SMTP_AUTH_TYPE = "oauth2"

# Standard settings
DEFAULT_EMAIL_PARAMS = {
    "mime_subtype": "mixed",
    "mime_charset": "utf-8",
    "files": None,
    "cc": None,
    "bcc": None,
    "custom_headers": None,
}

# DAG settings
TEST_DAG_ID = "test_dag"
TEST_TASK_ID = "test_task"
TEST_TASK_STATE = None
TEST_RUN_ID = "test_run"

# Jinja template patterns
DAG_ID_TEMPLATE_STRING = "{{dag.dag_id}}"
TI_TEMPLATE_STRING = "{{ti.task_id}}"

SENDER_EMAIL_SUFFIX = "sender@test.com"
RECEIVER_EMAIL_SUFFIX = "receiver@test.com"
# Base test values
TEST_SENDER = f"test_{SENDER_EMAIL_SUFFIX}"
TEST_RECEIVER = f"test_{RECEIVER_EMAIL_SUFFIX}"
TEST_SUBJECT = "subject"
TEST_BODY = "body"

NOTIFIER_DEFAULT_PARAMS = {
    "from_email": TEST_SENDER,
    "to": TEST_RECEIVER,
    "subject": TEST_SUBJECT,
    "html_content": TEST_BODY,
}


# Templated versions
@dataclass(frozen=True)
class TemplatedString:
    template: str

    @property
    def rendered(self) -> str:
        return self.template.replace(DAG_ID_TEMPLATE_STRING, TEST_DAG_ID).replace(
            TI_TEMPLATE_STRING, TEST_TASK_ID
        )


# DAG-based templates
TEMPLATED_SENDER = TemplatedString(f"{DAG_ID_TEMPLATE_STRING}_{SENDER_EMAIL_SUFFIX}")
TEMPLATED_RECEIVER = TemplatedString(f"{DAG_ID_TEMPLATE_STRING}_{RECEIVER_EMAIL_SUFFIX}")
TEMPLATED_SUBJECT = TemplatedString(f"{TEST_SUBJECT} {DAG_ID_TEMPLATE_STRING}")
TEMPLATED_BODY = TemplatedString(f"{TEST_BODY} {DAG_ID_TEMPLATE_STRING}")

# Task-based templates
TEMPLATED_TI_SUBJECT = TemplatedString(f"{TEST_SUBJECT} {TI_TEMPLATE_STRING}")
TEMPLATED_TI_SENDER = TemplatedString(f"{TI_TEMPLATE_STRING}_{SENDER_EMAIL_SUFFIX}")


class TestSmtpNotifier:
    @mock.patch("airflow.providers.smtp.notifications.smtp.SmtpHook")
    def test_notifier(_self, mock_smtphook_hook, create_dag_without_db):
        notifier = send_smtp_notification(**NOTIFIER_DEFAULT_PARAMS)
        notifier({"dag": create_dag_without_db(TEST_DAG_ID)})
        mock_smtphook_hook.return_value.__enter__().send_email_smtp.assert_called_once_with(
            **NOTIFIER_DEFAULT_PARAMS,
            smtp_conn_id=SMTP_CONN_ID,
            **DEFAULT_EMAIL_PARAMS,
        )

    @mock.patch("airflow.providers.smtp.notifications.smtp.SmtpHook")
    def test_notifier_with_notifier_class(self, mock_smtphook_hook, create_dag_without_db):
        notifier = SmtpNotifier(**NOTIFIER_DEFAULT_PARAMS)
        notifier({"dag": create_dag_without_db(TEST_DAG_ID)})
        mock_smtphook_hook.return_value.__enter__().send_email_smtp.assert_called_once_with(
            from_email=TEST_SENDER,
            to=TEST_RECEIVER,
            subject=TEST_SUBJECT,
            html_content=TEST_BODY,
            smtp_conn_id=SMTP_CONN_ID,
            **DEFAULT_EMAIL_PARAMS,
        )

    @mock.patch("airflow.providers.smtp.notifications.smtp.SmtpHook")
    def test_notifier_templated(self, mock_smtphook_hook, create_dag_without_db):
        notifier = SmtpNotifier(
            from_email=TEMPLATED_SENDER.template,
            to=TEMPLATED_RECEIVER.template,
            subject=TEMPLATED_SUBJECT.template,
            html_content=TEMPLATED_BODY.template,
        )
        context = {"dag": create_dag_without_db(TEST_DAG_ID)}

        notifier(context)

        mock_smtphook_hook.return_value.__enter__().send_email_smtp.assert_called_once_with(
            from_email=TEMPLATED_SENDER.rendered,
            to=TEMPLATED_RECEIVER.rendered,
            subject=TEMPLATED_SUBJECT.rendered,
            html_content=TEMPLATED_BODY.rendered,
            smtp_conn_id=SMTP_CONN_ID,
            **DEFAULT_EMAIL_PARAMS,
        )

    @mock.patch("airflow.providers.smtp.notifications.smtp.SmtpHook")
    def test_notifier_with_defaults(self, mock_smtphook_hook, create_dag_without_db, mock_task_instance):
        # TODO: we can use create_runtime_ti fixture in place of mock_task_instance once provider has minimum AF to Airflow 3.0+
        mock_ti = mock_task_instance(
            dag_id=TEST_DAG_ID,
            task_id=TEST_TASK_ID,
            run_id=TEST_RUN_ID,
            try_number=TRY_NUMBER,
            max_tries=0,
            state=TEST_TASK_STATE,
        )

        context = {"dag": create_dag_without_db(TEST_DAG_ID), "ti": mock_ti}
        notifier = SmtpNotifier(
            from_email=TEST_SENDER,
            to=TEST_RECEIVER,
        )
        mock_smtphook_hook.return_value.__enter__.return_value.subject_template = None
        mock_smtphook_hook.return_value.__enter__.return_value.html_content_template = None

        notifier(context)

        mock_smtphook_hook.return_value.__enter__().send_email_smtp.assert_called_once_with(
            from_email=TEST_SENDER,
            to=TEST_RECEIVER,
            subject=f"DAG {TEST_DAG_ID} - Task {TEST_TASK_ID} - Run ID {TEST_RUN_ID} in State {TEST_TASK_STATE}",
            html_content=mock.ANY,
            smtp_conn_id=SMTP_CONN_ID,
            **DEFAULT_EMAIL_PARAMS,
        )
        content = mock_smtphook_hook.return_value.__enter__().send_email_smtp.call_args.kwargs["html_content"]
        assert f"{TRY_NUMBER} of 1" in content

    @mock.patch("airflow.providers.smtp.notifications.smtp.SmtpHook")
    def test_notifier_with_nondefault_connection_extra(
        self, mock_smtphook_hook, create_dag_without_db, mock_task_instance
    ):
        # TODO: we can use create_runtime_ti fixture in place of mock_task_instance once provider has minimum AF to Airflow 3.0+
        ti = mock_task_instance(
            dag_id=TEST_DAG_ID,
            task_id=TEST_TASK_ID,
            run_id=TEST_RUN_ID,
            try_number=TRY_NUMBER,
            max_tries=0,
            state=TEST_TASK_STATE,
        )
        context = {"dag": create_dag_without_db(TEST_DAG_ID), "ti": ti}

        with (
            tempfile.NamedTemporaryFile(mode="wt", suffix=".txt") as f_subject,
            tempfile.NamedTemporaryFile(mode="wt", suffix=".txt") as f_content,
        ):
            f_subject.write(TEMPLATED_TI_SUBJECT.template)
            f_subject.flush()

            f_content.write(TEST_BODY)
            f_content.flush()
            mock_smtphook_hook.return_value.__enter__.return_value.from_email = TEMPLATED_TI_SENDER.template
            mock_smtphook_hook.return_value.__enter__.return_value.subject_template = f_subject.name
            mock_smtphook_hook.return_value.__enter__.return_value.html_content_template = f_content.name

            notifier = SmtpNotifier(to=TEST_RECEIVER)
            notifier(context)

            mock_smtphook_hook.return_value.__enter__().send_email_smtp.assert_called_once_with(
                from_email=TEMPLATED_TI_SENDER.rendered,
                to=TEST_RECEIVER,
                subject=TEMPLATED_TI_SUBJECT.rendered,
                html_content=TEST_BODY,
                smtp_conn_id=SMTP_CONN_ID,
                **DEFAULT_EMAIL_PARAMS,
            )

    @mock.patch("airflow.providers.smtp.notifications.smtp.SmtpHook")
    def test_notifier_oauth2_passes_auth_type(self, mock_smtphook_hook, create_dag_without_db):
        notifier = SmtpNotifier(**NOTIFIER_DEFAULT_PARAMS, auth_type=SMTP_AUTH_TYPE)

        notifier({"dag": create_dag_without_db(TEST_DAG_ID)})

        mock_smtphook_hook.assert_called_once_with(smtp_conn_id=SMTP_CONN_ID, auth_type=SMTP_AUTH_TYPE)


@pytest.mark.asyncio
@pytest.mark.skipif(not AIRFLOW_V_3_1_PLUS, reason="Async support was added to BaseNotifier in 3.1.0")
class TestSmtpNotifierAsync:
    @pytest.fixture
    def mock_smtp_client(self):
        """Create a mock SMTP object with async capabilities."""
        mock_smtp = AsyncMock()
        mock_smtp.asend_email_smtp = AsyncMock()
        return mock_smtp

    @pytest.fixture
    def mock_smtp_hook(self, mock_smtp_client):
        """Set up the SMTP hook with async context manager."""
        with mock.patch("airflow.providers.smtp.notifications.smtp.SmtpHook") as mock_hook:
            mock_hook.return_value.__aenter__ = AsyncMock(return_value=mock_smtp_client)
            yield mock_hook

    @pytest.mark.asyncio
    async def test_async_notifier(self, mock_smtp_hook, mock_smtp_client, create_dag_without_db):
        notifier = SmtpNotifier(
            **NOTIFIER_DEFAULT_PARAMS, context={"dag": create_dag_without_db(TEST_DAG_ID)}
        )
        await notifier.async_notify({"dag": create_dag_without_db(TEST_DAG_ID)})

        mock_smtp_client.asend_email_smtp.assert_called_once_with(
            smtp_conn_id=SMTP_CONN_ID,
            **NOTIFIER_DEFAULT_PARAMS,
            **DEFAULT_EMAIL_PARAMS,
        )

    async def test_async_notifier_with_notifier_class(
        self, mock_smtp_hook, mock_smtp_client, create_dag_without_db
    ):
        notifier = SmtpNotifier(
            **NOTIFIER_DEFAULT_PARAMS, context={"dag": create_dag_without_db(TEST_DAG_ID)}
        )

        await notifier

        mock_smtp_client.asend_email_smtp.assert_called_once_with(
            smtp_conn_id=SMTP_CONN_ID,
            **NOTIFIER_DEFAULT_PARAMS,
            **DEFAULT_EMAIL_PARAMS,
        )

    async def test_async_notifier_templated(self, mock_smtp_hook, mock_smtp_client, create_dag_without_db):
        notifier = SmtpNotifier(
            from_email=TEMPLATED_SENDER.template,
            to=TEMPLATED_RECEIVER.template,
            subject=TEMPLATED_SUBJECT.template,
            html_content=TEMPLATED_BODY.template,
            context={"dag": create_dag_without_db(TEST_DAG_ID)},
        )

        await notifier

        mock_smtp_client.asend_email_smtp.assert_called_once_with(
            smtp_conn_id=SMTP_CONN_ID,
            from_email=TEMPLATED_SENDER.rendered,
            to=TEMPLATED_RECEIVER.rendered,
            subject=TEMPLATED_SUBJECT.rendered,
            html_content=TEMPLATED_BODY.rendered,
            **DEFAULT_EMAIL_PARAMS,
        )

    async def test_async_notifier_with_defaults(
        self, mock_smtp_hook, mock_smtp_client, create_dag_without_db, mock_task_instance
    ):
        mock_smtp_client.subject_template = None
        mock_smtp_client.html_content_template = None
        mock_smtp_client.from_email = None

        notifier = SmtpNotifier(
            **NOTIFIER_DEFAULT_PARAMS, context={"dag": create_dag_without_db(TEST_DAG_ID)}
        )

        await notifier

        mock_smtp_client.asend_email_smtp.assert_called_once_with(
            smtp_conn_id=SMTP_CONN_ID,
            **NOTIFIER_DEFAULT_PARAMS,
            **DEFAULT_EMAIL_PARAMS,
        )

    async def test_async_notifier_with_nondefault_connection_extra(
        self, mock_smtp_hook, mock_smtp_client, create_dag_without_db, mock_task_instance
    ):
        with (
            tempfile.NamedTemporaryFile(mode="wt", suffix=".txt") as f_subject,
            tempfile.NamedTemporaryFile(mode="wt", suffix=".txt") as f_content,
        ):
            f_subject.write(TEST_SUBJECT)
            f_subject.flush()

            f_content.write(TEST_BODY)
            f_content.flush()

            mock_smtp_client.from_email = TEST_SENDER
            mock_smtp_client.subject_template = f_subject.name
            mock_smtp_client.html_content_template = f_content.name

            notifier = SmtpNotifier(to=TEST_RECEIVER, context={"dag": create_dag_without_db(TEST_DAG_ID)})

            await notifier

            mock_smtp_client.asend_email_smtp.assert_called_once_with(
                smtp_conn_id=SMTP_CONN_ID,
                from_email=TEST_SENDER,
                to=TEST_RECEIVER,
                subject=TEST_SUBJECT,
                html_content=TEST_BODY,
                **DEFAULT_EMAIL_PARAMS,
            )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import os
import smtplib
import tempfile
from email.mime.application import MIMEApplication
from unittest import mock
from unittest.mock import AsyncMock, Mock, call, patch

import aiosmtplib
import pytest

from airflow.exceptions import AirflowException
from airflow.providers.smtp.hooks.smtp import SmtpHook, build_xoauth2_string

from tests_common.test_utils.version_compat import AIRFLOW_V_3_1_PLUS

if AIRFLOW_V_3_1_PLUS:
    from airflow.sdk import Connection
else:
    from airflow.models.connection import Connection  # type: ignore[assignment]

smtplib_string = "airflow.providers.smtp.hooks.smtp.smtplib"

FROM_EMAIL = "from@example.com"
TO_EMAIL = "to@example.com"
TEST_EMAILS = [FROM_EMAIL, TO_EMAIL]

CONN_TYPE = "smtp"
SMTP_HOST = "smtp.example.com"
SMTP_LOGIN = "smtp_user"
SMTP_PASSWORD = "smtp_password"
ACCESS_TOKEN = "test-token"

CONN_ID_DEFAULT = "smtp_default"
CONN_ID_NONSSL = "smtp_nonssl"
CONN_ID_SSL_EXTRA = "smtp_ssl_extra"
CONN_ID_OAUTH = "smtp_oauth2"

DEFAULT_PORT = 465
NONSSL_PORT = 587

DEFAULT_TIMEOUT = 30
DEFAULT_RETRY_LIMIT = 5

TEST_SUBJECT = "test subject"
TEST_BODY = "<html>Test</html>"

SERVER_DISCONNECTED_ERROR = aiosmtplib.errors.SMTPServerDisconnected("Server disconnected")


def _create_fake_smtp(mock_smtplib, use_ssl=True):
    if use_ssl:
        mock_conn = Mock(spec=smtplib.SMTP_SSL)
        mock_smtplib.SMTP_SSL.return_value = mock_conn
    else:
        mock_conn = Mock(spec=smtplib.SMTP)
        mock_smtplib.SMTP.return_value = mock_conn

    mock_conn.close.return_value = ("OK", [])

    return mock_conn


class TestSmtpHook:
    @pytest.fixture(autouse=True)
    def setup_connections(self, create_connection_without_db):
        create_connection_without_db(
            Connection(
                conn_id=CONN_ID_DEFAULT,
                conn_type=CONN_TYPE,
                host=SMTP_HOST,
                login=SMTP_LOGIN,
                password=SMTP_PASSWORD,
                port=DEFAULT_PORT,
                extra=json.dumps(dict(from_email=FROM_EMAIL)),
            )
        )
        create_connection_without_db(
            Connection(
                conn_id=CONN_ID_NONSSL,
                conn_type=CONN_TYPE,
                host=SMTP_HOST,
                login=SMTP_LOGIN,
                password=SMTP_PASSWORD,
                port=NONSSL_PORT,
                extra=json.dumps(dict(disable_ssl=True, from_email=FROM_EMAIL)),
            )
        )
        create_connection_without_db(
            Connection(
                conn_id=CONN_ID_OAUTH,
                conn_type=CONN_TYPE,
                host=SMTP_HOST,
                login=SMTP_LOGIN,
                password=SMTP_PASSWORD,
                port=NONSSL_PORT,
                extra=json.dumps(dict(disable_ssl=True, from_email=FROM_EMAIL, access_token=ACCESS_TOKEN)),
            )
        )
        create_connection_without_db(
            Connection(
                conn_id=CONN_ID_SSL_EXTRA,
                conn_type=CONN_TYPE,
                host=SMTP_HOST,
                login=None,
                password="None",
                port=DEFAULT_PORT,
                extra=json.dumps(dict(use_ssl=True, ssl_context="none", from_email=FROM_EMAIL)),
            )
        )

    @pytest.mark.parametrize(
        "conn_id, use_ssl, expected_port, create_context",
        [
            pytest.param(CONN_ID_DEFAULT, True, DEFAULT_PORT, True, id="ssl-connection"),
            pytest.param(CONN_ID_NONSSL, False, NONSSL_PORT, False, id="non-ssl-connection"),
        ],
    )
    @patch(smtplib_string)
    @patch("ssl.create_default_context")
    def test_connect_and_disconnect(
        self, create_default_context, mock_smtplib, conn_id, use_ssl, expected_port, create_context
    ):
        """Test sync connection with different configurations."""
        mock_conn = _create_fake_smtp(mock_smtplib, use_ssl=use_ssl)

        with SmtpHook(smtp_conn_id=conn_id):
            pass

        if create_context:
            assert create_default_context.called
            mock_smtplib.SMTP_SSL.assert_called_once_with(
                host=SMTP_HOST,
                port=expected_port,
                timeout=DEFAULT_TIMEOUT,
                context=create_default_context.return_value,
            )
        else:
            mock_smtplib.SMTP.assert_called_once_with(
                host=SMTP_HOST,
                port=expected_port,
                timeout=DEFAULT_TIMEOUT,
            )

        mock_conn.login.assert_called_once_with(SMTP_LOGIN, SMTP_PASSWORD)
        assert mock_conn.close.call_count == 1

    @patch(smtplib_string)
    def test_get_email_address_single_email(self, mock_smtplib):
        with SmtpHook() as smtp_hook:
            assert smtp_hook._get_email_address_list(FROM_EMAIL) == [FROM_EMAIL]

    @pytest.mark.parametrize(
        "email_input",
        [
            pytest.param(f"{FROM_EMAIL}, {TO_EMAIL}", id="comma_separated"),
            pytest.param(f"{FROM_EMAIL}; {TO_EMAIL}", id="semicolon_separated"),
            pytest.param([FROM_EMAIL, TO_EMAIL], id="list_input"),
            pytest.param((FROM_EMAIL, TO_EMAIL), id="tuple_input"),
        ],
    )
    @patch(smtplib_string)
    def test_get_email_address_parsing(self, mock_smtplib, email_input):
        with SmtpHook() as smtp_hook:
            assert smtp_hook._get_email_address_list(email_input) == TEST_EMAILS

    @pytest.mark.parametrize(
        "invalid_input",
        [
            pytest.param(1, id="invalid_scalar_type"),
            pytest.param([FROM_EMAIL, 2], id="invalid_type_in_list"),
        ],
    )
    @patch(smtplib_string)
    def test_get_email_address_invalid_types(self, mock_smtplib, invalid_input):
        with pytest.raises(TypeError):
            with SmtpHook() as smtp_hook:
                smtp_hook._get_email_address_list(invalid_input)

    @patch(smtplib_string)
    def test_build_mime_message(self, mock_smtplib):
        mail_from = FROM_EMAIL
        mail_to = TO_EMAIL
        subject = TEST_SUBJECT
        html_content = TEST_BODY
        custom_headers = {"Reply-To": "reply_to@example.com"}
        with SmtpHook() as smtp_hook:
            msg, recipients = smtp_hook._build_mime_message(
                mail_from=mail_from,
                to=mail_to,
                subject=subject,
                html_content=html_content,
                custom_headers=custom_headers,
            )

        assert "From" in msg
        assert "To" in msg
        assert "Subject" in msg
        assert "Reply-To" in msg
        assert [mail_to] == recipients
        assert msg["To"] == ",".join(recipients)

    @patch(smtplib_string)
    def test_send_smtp(self, mock_smtplib):
        mock_send_mime = mock_smtplib.SMTP_SSL().sendmail
        with SmtpHook() as smtp_hook, tempfile.NamedTemporaryFile() as attachment:
            attachment.write(b"attachment")
            attachment.seek(0)
            smtp_hook.send_email_smtp(
                to=TO_EMAIL, subject=TEST_SUBJECT, html_content=TEST_BODY, files=[attachment.name]
            )
            assert mock_send_mime.called
            _, call_args = mock_send_mime.call_args
            assert call_args["from_addr"] == FROM_EMAIL
            assert call_args["to_addrs"] == [TO_EMAIL]
            msg = call_args["msg"]
            assert f"Subject: {TEST_SUBJECT}" in msg
            assert f"From: {FROM_EMAIL}" in msg
            filename = 'attachment; filename="' + os.path.basename(attachment.name) + '"'
            assert filename in msg
            mimeapp = MIMEApplication("attachment")
            assert mimeapp.get_payload() in msg

    @patch("airflow.providers.smtp.hooks.smtp.SmtpHook.get_connection")
    @patch(smtplib_string)
    def test_hook_conn(self, mock_smtplib, mock_hook_conn):
        mock_conn = Mock()
        mock_conn.login = SMTP_LOGIN
        mock_conn.password = SMTP_PASSWORD
        mock_conn.extra_dejson = {"disable_ssl": False}
        mock_hook_conn.return_value = mock_conn
        smtp_client_mock = mock_smtplib.SMTP_SSL()
        with SmtpHook() as smtp_hook:
            smtp_hook.send_email_smtp(
                to=TO_EMAIL,
                subject=TEST_SUBJECT,
                html_content=TEST_BODY,
                from_email=FROM_EMAIL,
            )

            mock_hook_conn.assert_called_with(CONN_ID_DEFAULT)
            smtp_client_mock.login.assert_called_once_with(SMTP_LOGIN, SMTP_PASSWORD)
            smtp_client_mock.sendmail.assert_called_once()
        assert smtp_client_mock.close.called

    @pytest.mark.parametrize(
        "conn_id, ssl_context, create_context_called, use_default_context",
        [
            pytest.param(CONN_ID_DEFAULT, "default", True, True, id="default_context"),
            pytest.param(CONN_ID_SSL_EXTRA, "none", False, False, id="none_context"),
            pytest.param(CONN_ID_DEFAULT, "default", True, True, id="explicit_default_context"),
        ],
    )
    @patch("smtplib.SMTP_SSL")
    @patch("smtplib.SMTP")
    @patch("ssl.create_default_context")
    def test_send_mime_ssl_context(
        self,
        create_default_context,
        mock_smtp,
        mock_smtp_ssl,
        conn_id,
        ssl_context,
        create_context_called,
        use_default_context,
    ):
        mock_smtp_ssl.return_value = Mock()

        with SmtpHook(conn_id) as smtp_hook:
            smtp_hook.send_email_smtp(
                to=TO_EMAIL, subject=TEST_SUBJECT, html_content=TEST_BODY, from_email=FROM_EMAIL
            )

        assert not mock_smtp.called
        if use_default_context:
            assert create_default_context.called
            expected_context = create_default_context.return_value
        else:
            create_default_context.assert_not_called()
            expected_context = None

        mock_smtp_ssl.assert_called_once_with(
            host=SMTP_HOST, port=DEFAULT_PORT, timeout=DEFAULT_TIMEOUT, context=expected_context
        )

    @patch("smtplib.SMTP_SSL")
    @patch("smtplib.SMTP")
    @patch("ssl.create_default_context")
    def test_send_mime_ssl(self, create_default_context, mock_smtp, mock_smtp_ssl):
        mock_smtp_ssl.return_value = Mock()

        with SmtpHook() as smtp_hook:
            smtp_hook.send_email_smtp(
                to=TO_EMAIL,
                subject=TEST_SUBJECT,
                html_content=TEST_BODY,
                from_email=FROM_EMAIL,
            )

        assert not mock_smtp.called
        assert create_default_context.called
        mock_smtp_ssl.assert_called_once_with(
            host=SMTP_HOST,
            port=DEFAULT_PORT,
            timeout=DEFAULT_TIMEOUT,
            context=create_default_context.return_value,
        )

    @patch("smtplib.SMTP_SSL")
    @patch("smtplib.SMTP")
    def test_send_mime_nossl(self, mock_smtp, mock_smtp_ssl):
        mock_smtp.return_value = Mock()

        with SmtpHook(smtp_conn_id=CONN_ID_NONSSL) as smtp_hook:
            smtp_hook.send_email_smtp(
                to=TO_EMAIL, subject=TEST_SUBJECT, html_content=TEST_BODY, from_email=FROM_EMAIL
            )

        assert not mock_smtp_ssl.called
        mock_smtp.assert_called_once_with(host=SMTP_HOST, port=NONSSL_PORT, timeout=DEFAULT_TIMEOUT)

    @patch("smtplib.SMTP")
    def test_send_mime_noauth(self, mock_smtp, create_connection_without_db):
        mock_smtp.return_value = Mock()
        conn = Connection(
            conn_id="smtp_noauth",
            conn_type=CONN_TYPE,
            host=SMTP_HOST,
            login=None,
            password="None",
            port=NONSSL_PORT,
            extra=json.dumps(dict(disable_ssl=True, from_email=FROM_EMAIL)),
        )
        create_connection_without_db(conn)
        with SmtpHook(smtp_conn_id="smtp_noauth") as smtp_hook:
            smtp_hook.send_email_smtp(
                to=TO_EMAIL,
                subject=TEST_SUBJECT,
                html_content=TEST_BODY,
                from_email=FROM_EMAIL,
            )
        mock_smtp.assert_called_once_with(host=SMTP_HOST, port=NONSSL_PORT, timeout=DEFAULT_TIMEOUT)
        assert not mock_smtp.login.called

    @patch("smtplib.SMTP_SSL")
    @patch("smtplib.SMTP")
    def test_send_mime_dryrun(self, mock_smtp, mock_smtp_ssl):
        with SmtpHook() as smtp_hook:
            smtp_hook.send_email_smtp(
                to=TO_EMAIL,
                subject=TEST_SUBJECT,
                html_content=TEST_BODY,
                dryrun=True,
            )

        assert not mock_smtp.sendmail.called
        assert not mock_smtp_ssl.sendmail.called

    @patch("smtplib.SMTP_SSL")
    def test_send_mime_ssl_complete_failure(self, mock_smtp_ssl):
        mock_smtp_ssl().sendmail.side_effect = smtplib.SMTPServerDisconnected()

        with SmtpHook() as smtp_hook:
            with pytest.raises(smtplib.SMTPServerDisconnected):
                smtp_hook.send_email_smtp(
                    to=TO_EMAIL,
                    subject=TEST_SUBJECT,
                    html_content=TEST_BODY,
                )

        assert mock_smtp_ssl().sendmail.call_count == DEFAULT_RETRY_LIMIT

    @patch("email.message.Message.as_string")
    @patch("smtplib.SMTP_SSL")
    def test_send_mime_partial_failure(self, mock_smtp_ssl, mime_message_mock):
        mime_message_mock.return_value = "msg"
        final_mock = Mock()
        side_effects = [smtplib.SMTPServerDisconnected(), smtplib.SMTPServerDisconnected(), final_mock]
        mock_smtp_ssl.side_effect = side_effects
        with SmtpHook() as smtp_hook:
            smtp_hook.send_email_smtp(
                to=TO_EMAIL,
                subject=TEST_SUBJECT,
                html_content=TEST_BODY,
            )
        assert mock_smtp_ssl.call_count == side_effects.index(final_mock) + 1
        assert final_mock.starttls.called
        final_mock.sendmail.assert_called_once_with(from_addr=FROM_EMAIL, to_addrs=[TO_EMAIL], msg="msg")
        assert final_mock.close.called

    @patch("smtplib.SMTP_SSL")
    @patch("ssl.create_default_context")
    def test_send_mime_custom_timeout_retrylimit(
        self, create_default_context, mock_smtp_ssl, create_connection_without_db
    ):
        mock_smtp_ssl().sendmail.side_effect = smtplib.SMTPServerDisconnected()
        custom_retry_limit = 10
        custom_timeout = 60
        fake_conn = Connection(
            conn_id="mock_conn",
            conn_type=CONN_TYPE,
            host=SMTP_HOST,
            login=SMTP_LOGIN,
            password=SMTP_PASSWORD,
            port=DEFAULT_PORT,
            extra=json.dumps(
                dict(from_email=FROM_EMAIL, timeout=custom_timeout, retry_limit=custom_retry_limit)
            ),
        )
        create_connection_without_db(fake_conn)

        with SmtpHook(smtp_conn_id="mock_conn") as smtp_hook:
            with pytest.raises(smtplib.SMTPServerDisconnected):
                smtp_hook.send_email_smtp(to=TO_EMAIL, subject=TEST_SUBJECT, html_content=TEST_BODY)

        expected_call = call(
            host=fake_conn.host,
            port=fake_conn.port,
            timeout=fake_conn.extra_dejson["timeout"],
            context=create_default_context.return_value,
        )
        assert expected_call in mock_smtp_ssl.call_args_list
        assert create_default_context.called
        assert mock_smtp_ssl().sendmail.call_count == 10

    @patch(smtplib_string)
    def test_oauth2_auth_called(self, mock_smtplib):
        mock_conn = _create_fake_smtp(mock_smtplib, use_ssl=False)

        with SmtpHook(smtp_conn_id=CONN_ID_OAUTH, auth_type="oauth2") as smtp_hook:
            smtp_hook.send_email_smtp(
                to=TO_EMAIL,
                subject=TEST_SUBJECT,
                html_content=TEST_BODY,
                from_email=FROM_EMAIL,
            )

        assert mock_conn.auth.called
        args, _ = mock_conn.auth.call_args
        assert args[0] == "XOAUTH2"
        assert build_xoauth2_string(SMTP_LOGIN, ACCESS_TOKEN) == args[1]()

    @patch(smtplib_string)
    def test_oauth2_missing_token_raises(self, mock_smtplib, create_connection_without_db):
        mock_conn = _create_fake_smtp(mock_smtplib, use_ssl=False)

        create_connection_without_db(
            Connection(
                conn_id="smtp_oauth2_empty",
                conn_type=CONN_TYPE,
                host=SMTP_HOST,
                login=SMTP_LOGIN,
                password=SMTP_PASSWORD,
                port=NONSSL_PORT,
                extra=json.dumps(dict(disable_ssl=True, from_email=FROM_EMAIL)),
            )
        )

        with pytest.raises(AirflowException):
            with SmtpHook(smtp_conn_id="smtp_oauth2_empty", auth_type="oauth2") as h:
                h.send_email_smtp(
                    to=TO_EMAIL,
                    subject=TEST_SUBJECT,
                    html_content=TEST_BODY,
                    from_email=FROM_EMAIL,
                )

        assert not mock_conn.auth.called


@pytest.mark.asyncio
@pytest.mark.skipif(not AIRFLOW_V_3_1_PLUS, reason="Async support was added to BaseNotifier in 3.1.0")
class TestSmtpHookAsync:
    """Tests for async functionality in SmtpHook."""

    @pytest.fixture(autouse=True)
    def setup_connections(self, create_connection_without_db):
        create_connection_without_db(
            Connection(
                conn_id=CONN_ID_DEFAULT,
                conn_type=CONN_TYPE,
                host=SMTP_HOST,
                login=SMTP_LOGIN,
                password=SMTP_PASSWORD,
                port=DEFAULT_PORT,
                extra=json.dumps(dict(from_email=FROM_EMAIL, ssl_context="default")),
            )
        )
        create_connection_without_db(
            Connection(
                conn_id=CONN_ID_NONSSL,
                conn_type=CONN_TYPE,
                host=SMTP_HOST,
                login=SMTP_LOGIN,
                password=SMTP_PASSWORD,
                port=NONSSL_PORT,
                extra=json.dumps(dict(disable_ssl=True, from_email=FROM_EMAIL)),
            )
        )

    @pytest.fixture
    def mock_smtp_client(self):
        """Create a mock SMTP client with async capabilities."""
        mock_client = AsyncMock(spec=aiosmtplib.SMTP)
        mock_client.starttls = AsyncMock()
        mock_client.auth_login = AsyncMock()
        mock_client.sendmail = AsyncMock()
        mock_client.quit = AsyncMock()
        return mock_client

    @pytest.fixture
    def mock_smtp(self, mock_smtp_client):
        """Set up the SMTP mock with context manager."""
        with mock.patch("airflow.providers.smtp.hooks.smtp.aiosmtplib.SMTP") as mock_smtp:
            mock_smtp.return_value = mock_smtp_client
            yield mock_smtp

    @pytest.fixture
    def mock_get_connection(self):
        """Mock the async connection retrieval."""
        with mock.patch("airflow.sdk.bases.hook.BaseHook.aget_connection") as mock_conn:

            async def async_get_connection(conn_id):
                from airflow.sdk.definitions.connection import Connection

                return Connection.from_json(os.environ[f"AIRFLOW_CONN_{conn_id.upper()}"])

            mock_conn.side_effect = async_get_connection
            yield mock_conn

    @staticmethod
    def _create_fake_async_smtp(mock_smtp):
        mock_client = AsyncMock(spec=aiosmtplib.SMTP)
        mock_client.starttls = AsyncMock()
        mock_client.auth_login = AsyncMock()
        mock_client.sendmail = AsyncMock()
        mock_client.quit = AsyncMock()
        mock_smtp.return_value = mock_client
        return mock_client

    @pytest.mark.parametrize(
        "conn_id, expected_port, expected_ssl",
        [
            pytest.param(CONN_ID_NONSSL, NONSSL_PORT, False, id="non-ssl-connection"),
            pytest.param(CONN_ID_DEFAULT, DEFAULT_PORT, True, id="ssl-connection"),
        ],
    )
    async def test_async_connection(
        self, mock_smtp, mock_smtp_client, mock_get_connection, conn_id, expected_port, expected_ssl
    ):
        """Test async connection with different configurations."""
        async with SmtpHook(smtp_conn_id=conn_id) as hook:
            assert hook is not None

        mock_smtp.assert_called_once_with(
            hostname=SMTP_HOST,
            port=expected_port,
            timeout=DEFAULT_TIMEOUT,
            use_tls=expected_ssl,
            start_tls=None if expected_ssl else True,
        )

        if expected_ssl:
            assert mock_smtp_client.starttls.await_count == 1

        assert mock_smtp_client.auth_login.await_count == 1
        mock_smtp_client.auth_login.assert_awaited_once_with(SMTP_LOGIN, SMTP_PASSWORD)

    @pytest.mark.asyncio
    async def test_async_send_email(self, mock_smtp, mock_smtp_client, mock_get_connection):
        """Test async email sending functionality."""
        async with SmtpHook() as hook:
            await hook.asend_email_smtp(
                to=TO_EMAIL,
                subject=TEST_SUBJECT,
                html_content=TEST_BODY,
            )

        assert mock_smtp_client.sendmail.called
        #  The async version of sendmail only supports positional arguments
        #  for some reason, so we have to check these by positional args
        call_args = mock_smtp_client.sendmail.await_args.args
        assert call_args[0] == FROM_EMAIL  # sender is first positional arg
        assert call_args[1] == [TO_EMAIL]  # recipients is the second positional arg
        assert f"Subject: {TEST_SUBJECT}" in call_args[2]  # message is the third positional arg

    @pytest.mark.parametrize(
        "side_effect, expected_calls, should_raise",
        [
            pytest.param(
                [SERVER_DISCONNECTED_ERROR, SERVER_DISCONNECTED_ERROR, None],
                3,
                False,
                id="success_after_retries",
            ),
            pytest.param(SERVER_DISCONNECTED_ERROR, DEFAULT_RETRY_LIMIT, True, id="max_retries_exceeded"),
        ],
    )
    @pytest.mark.asyncio
    async def test_async_send_email_retries(
        self, mock_smtp, mock_smtp_client, mock_get_connection, side_effect, expected_calls, should_raise
    ):
        mock_smtp_client.sendmail.side_effect = side_effect

        if should_raise:
            with pytest.raises(aiosmtplib.errors.SMTPServerDisconnected):
                async with SmtpHook() as hook:
                    await hook.asend_email_smtp(
                        to=TO_EMAIL,
                        subject=TEST_SUBJECT,
                        html_content=TEST_BODY,
                    )
        else:
            async with SmtpHook() as hook:
                await hook.asend_email_smtp(
                    to=TO_EMAIL,
                    subject=TEST_SUBJECT,
                    html_content=TEST_BODY,
                )

        assert mock_smtp_client.sendmail.call_count == expected_calls

    async def test_async_send_email_dryrun(self, mock_smtp, mock_smtp_client, mock_get_connection):
        """Test async email sending in dryrun mode."""
        async with SmtpHook() as hook:
            await hook.asend_email_smtp(
                to=TO_EMAIL,
                subject=TEST_SUBJECT,
                html_content=TEST_BODY,
                dryrun=True,
            )

        mock_smtp_client.sendmail.assert_not_awaited()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import base64
import json
import tempfile
from unittest.mock import MagicMock, patch

from airflow.models import Connection
from airflow.providers.smtp.operators.smtp import EmailOperator

smtplib_string = "airflow.providers.smtp.hooks.smtp.smtplib"


class TestEmailOperator:
    def setup_method(self):
        self.default_op_kwargs = dict(to="to", subject="subject", html_content="content")

    @patch("airflow.providers.smtp.hooks.smtp.SmtpHook.get_connection")
    @patch(smtplib_string)
    def test_loading_sender_email_from_connection(self, mock_smtplib, mock_hook_conn):
        """Check if the EmailOperator is able to load the sender email from the smtp connection."""
        custom_retry_limit = 10
        custom_timeout = 60
        sender_email = "sender_email"
        with (
            tempfile.NamedTemporaryFile(mode="wt", suffix=".txt") as f_subject,
            tempfile.NamedTemporaryFile(mode="wt", suffix=".txt") as f_content,
        ):
            f_subject.write("Task {{ ti.task_id }} failed")
            f_subject.flush()

            f_content.write("Mock content goes here")
            f_content.flush()

            mock_hook_conn.return_value = Connection(
                conn_id="mock_conn",
                conn_type="smtp",
                host="smtp_server_address",
                login="smtp_user",
                password="smtp_password",
                port=465,
                extra=json.dumps(
                    dict(
                        from_email=sender_email,
                        timeout=custom_timeout,
                        retry_limit=custom_retry_limit,
                        subject_template=f_subject.name,
                        html_content_template=f_content.name,
                    )
                ),
            )
            smtp_client_mock = mock_smtplib.SMTP_SSL()
            op = EmailOperator(task_id="test_email", to="to")
            op.execute({"ti": MagicMock(task_id="some_id")})
            call_args = smtp_client_mock.sendmail.call_args.kwargs
            assert call_args["from_addr"] == sender_email
            assert "Subject: Task some_id failed" in call_args["msg"]
            assert (
                base64.b64encode("Mock content goes here".encode("ascii")).decode("ascii") in call_args["msg"]
            )

    def test_assert_templated_fields(self):
        """Test expected templated fields."""
        operator = EmailOperator(task_id="test_assert_templated_fields", **self.default_op_kwargs)
        template_fields = ("to", "from_email", "subject", "html_content", "files", "cc", "bcc")
        assert operator.template_fields == template_fields



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any

from airflow.exceptions import AirflowException
from airflow.providers.smtp.hooks.smtp import SmtpHook
from airflow.providers.smtp.version_compat import BaseOperator

if TYPE_CHECKING:
    from airflow.providers.smtp.version_compat import Context


class EmailOperator(BaseOperator):
    """
    Sends an email.

    :param to: list of emails to send the email to. (templated)
    :param from_email: email to send from. (templated)
    :param subject: subject line for the email. (templated)
    :param html_content: content of the email, html markup
        is allowed. (templated)
    :param files: file names to attach in email (templated)
    :param cc: list of recipients to be added in CC field (templated)
    :param bcc: list of recipients to be added in BCC field (templated)
    :param mime_subtype: MIME sub content type
    :param mime_charset: character set parameter added to the Content-Type
        header.
    :param custom_headers: additional headers to add to the MIME message.
    """

    template_fields: Sequence[str] = ("to", "from_email", "subject", "html_content", "files", "cc", "bcc")
    template_fields_renderers = {"html_content": "html"}
    template_ext: Sequence[str] = (".html",)
    ui_color = "#e6faf9"

    def __init__(
        self,
        *,
        to: list[str] | str,
        subject: str | None = None,
        html_content: str | None = None,
        from_email: str | None = None,
        files: list | None = None,
        cc: list[str] | str | None = None,
        bcc: list[str] | str | None = None,
        mime_subtype: str = "mixed",
        mime_charset: str = "utf-8",
        conn_id: str = "smtp_default",
        custom_headers: dict[str, Any] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.to = to
        self.subject = subject
        self.html_content = html_content
        self.from_email = from_email
        self.files = files or []
        self.cc = cc
        self.bcc = bcc
        self.mime_subtype = mime_subtype
        self.mime_charset = mime_charset
        self.conn_id = conn_id
        self.custom_headers = custom_headers

    @staticmethod
    def _read_template(template_path: str) -> str:
        return Path(template_path).read_text().replace("\n", "").strip()

    def execute(self, context: Context):
        with SmtpHook(smtp_conn_id=self.conn_id) as smtp_hook:
            fields_to_re_render = []
            if self.from_email is None:
                if smtp_hook.from_email is None:
                    raise AirflowException("You should provide `from_email` or define it in the connection.")
                self.from_email = smtp_hook.from_email
                fields_to_re_render.append("from_email")
            if self.subject is None:
                if smtp_hook.subject_template is None:
                    raise AirflowException(
                        "You should provide `subject` or define `subject_template` in the connection."
                    )
                self.subject = self._read_template(smtp_hook.subject_template)
                fields_to_re_render.append("subject")
            if self.html_content is None:
                if smtp_hook.html_content_template is None:
                    raise AirflowException(
                        "You should provide `html_content` or define `html_content_template` in the connection."
                    )
                self.html_content = self._read_template(smtp_hook.html_content_template)
                fields_to_re_render.append("html_content")
            if fields_to_re_render:
                self._do_render_template_fields(
                    self, fields_to_re_render, context, self.get_template_env(), set()
                )
            return smtp_hook.send_email_smtp(
                to=self.to,
                subject=self.subject,
                html_content=self.html_content,
                from_email=self.from_email,
                files=self.files,
                cc=self.cc,
                bcc=self.bcc,
                mime_subtype=self.mime_subtype,
                mime_charset=self.mime_charset,
                conn_id=self.conn_id,
                custom_headers=self.custom_headers,
            )
